/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */



/******************* Operating System Interface Routines *******************\
*                                                                           *
* Module: interface.c                                                       *
*   This is the OS interface module.  All operating system transactions     *
*   pass through these routines.  No other operating system specific code   *
*   or data should exist in the source.                                     *
*                                                                           *
\***************************************************************************/

#define  __NO_VERSION__
#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"


BOOL os_is_administrator(PHWINFO pDev)
{
    return suser();
}

U032 os_get_page_size(VOID)
{
    return PAGE_SIZE;
}

U032 os_get_page_mask(VOID)
{
    return PAGE_MASK;
}

U008* os_string_copy(
    U008 *dst,
    const U008 *src
)
{
    return strcpy(dst, src);
}

RM_STATUS os_strncpy_from_user(
    U008 *dst,
    const U008 *src,
    U032 n
)
{
#if defined(NVCPU_IA64)
    return copy_from_user(dst, src, n) ? RM_ERR_BAD_ADDRESS : RM_OK;
#else
    return strncpy_from_user(dst, src, n) ? RM_ERR_BAD_ADDRESS : RM_OK;
#endif
}

S032 os_string_compare(
    const U008 *s1,
    const U008 *s2
)
{
    return strcmp(s1, s2);
}

U032 os_string_length(const U008* str)
{
    return strlen(str);
}

U008* os_mem_copy(
    U008 *dst,
    const U008 *src,
    U032 length
)
{
    return memcpy(dst, src, length);
}

RM_STATUS os_memcpy_from_user(
    VOID *dst,
    const VOID* src,
    U032 length
)
{
    return copy_from_user(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

RM_STATUS os_memcpy_to_user(
    VOID *dst,
    const VOID* src,
    U032 length
)
{
    return copy_to_user(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

VOID* os_mem_set(
    VOID* dst,
    U008 c,
    U032 length
)
{
    return memset(dst, c, length);
}

S032 os_mem_cmp(
    const U008 *buf0,
    const U008* buf1,
    U032 length
)
{
    return memcmp(buf0, buf1, length);
}

VOID* os_copy_in_ioctl_param(
    VOID *dst,
    VOID *src,
    U032 length
)
{
    if (copy_from_user(dst, src, length) > 0) {
        return NULL;
    } else {
        return dst;
    }
}

VOID* os_copy_out_ioctl_param(
    VOID *dst,
    VOID *src,
    U032 length
)
{
    if (copy_to_user(dst, src, length) > 0) {
        return NULL;
    } else {
        return (VOID *)1;
    }
}

/*
 * Operating System Memory Functions
 *
 * There are 2 interesting aspects of resource manager memory allocations
 * that need special consideration on Linux:
 *
 * 1. They are typically very large, (e.g. single allocations of 164KB)
 *
 * 2. The resource manager assumes that it can safely allocate memory in
 *    interrupt handlers.
 *
 * The first requires that we call vmalloc, the second kmalloc. We decide
 * which one to use at run time, based on the size of the request and the
 * context. Allocations larger than 128KB require vmalloc, in the context
 * of an ISR they fail.
 */

#define KMALLOC_LIMIT 131072

RM_STATUS os_alloc_mem(
    VOID **address,
    U032 size
)
{
    if (in_interrupt()) {
        if (size <= KMALLOC_LIMIT) {
            /*
             * The allocation request can be serviced with the Linux slab
             * allocator; the likelyhood of failure is still fairly high,
             * since kmalloc can't sleep.
             */
            *address = kmalloc(size, GFP_ATOMIC);
        } else {
            /*
             * If the requested amount of memory exceeds kmalloc's limit,
             * we can't fulfill the request, as vmalloc can not be called
             * from an interrupt context (bottom-half, ISR).
             */
            *address = NULL;
        }
    } else {
        if (size <= KMALLOC_LIMIT)
            *address = kmalloc(size, GFP_KERNEL);
        else
            *address = vmalloc(size);
    }

    return *address ? RM_OK : RM_ERR_NO_FREE_MEM;
}

void os_free_mem(VOID *address)
{
    unsigned long va = (unsigned long) address;

    if (va >= VMALLOC_START && va < VMALLOC_END)
        vfree(address);
    else
        kfree(address);
}

static inline int __get_order(U032 count)
{
    int order = 0;
    while ((1 << order) < count) order++;
    return order;
}

/* note that since contiguous pages are allocated by order
 * we may allocate more pages than the caller really asked for.
 * we'll only lock down the number of pages the caller asked for.
 */
RM_STATUS os_alloc_contig_pages(
    VOID **address,
    U032 size
)
{
    int i;
    unsigned long va;

    /*
     * The size argument is assumed to be a multiple of OS_PAGE_SIZE, and
     * can thus be easily translated into a page count, and then into the
     * order argument expected by __get_free_pages - we may allocate more
     * memory than requested.
     */
    int count = size / PAGE_SIZE;

    va = __get_free_pages(NV_GFP_HW, __get_order(count));
    if (!va)
        return RM_ERR_NO_FREE_MEM;

    *address = (void *) va;

    for (i = 0; i < count; i++) {
        mem_map_reserve(GET_MAP_NR(__pa((va))));
        va += PAGE_SIZE;
    }

    return RM_OK;
}

/* we may have actually allocated more pages than the caller requested,
 * due to contiguous allocs being by order. Note though, that we only
 * locked down the number of pages the caller requested to be allocated.
 * we get the size the caller requested, so use that value to unlock
 * the same number of pages we locked, and to calculate the order we
 * allocated, so we properly free the allocation.
 */
VOID os_free_contig_pages(
    VOID *address,
    U032 size
)
{
    int i;
    int count = size / PAGE_SIZE;
    unsigned long va = (unsigned long) address;

    for (i = 0; i < count; i++) {
        mem_map_unreserve(GET_MAP_NR(__pa((va))));
        va += PAGE_SIZE;
    }

    va = (unsigned long) address;
    free_pages(va, __get_order(count));
}


/*****************************************************************************
*
*   Name: osGetCurrentTime
*
*****************************************************************************/

RM_STATUS os_get_current_time(
    U032 *seconds,
    U032 *useconds
)
{
    struct timeval tm;

    do_gettimeofday(&tm);

    *seconds = tm.tv_sec;
    *useconds = tm.tv_usec;

    return RM_OK;
}

//---------------------------------------------------------------------------
//
//  Misc services.
//
//---------------------------------------------------------------------------

#define NV_MAX_ISR_UDELAY           20000
#define NV_MAX_ISR_MDELAY           (NV_MAX_ISR_UDELAY / 1000)
#define NV_MSECS_PER_JIFFIE         (1000 / HZ)
#define NV_MSECS_TO_JIFFIES(msec)   ((msec) * HZ / 1000)

/*
 * It is generally a bad idea to use udelay() to wait for more than
 * a few milliseconds. Since the caller is most likely not aware of
 * this, we use mdelay() for any full millisecond to be safe.
 */

RM_STATUS os_delay_us(U032 MicroSeconds)
{
    unsigned long mdelay_safe_msec;
    unsigned long usec;

    if (in_irq() && MicroSeconds > NV_MAX_ISR_UDELAY)
        return RM_ERROR;
    
    mdelay_safe_msec = MicroSeconds / 1000;
    if (mdelay_safe_msec)
        mdelay(mdelay_safe_msec);

    usec = MicroSeconds % 1000;
    if (usec)
        udelay(usec);

    return RM_OK;
}

/* 
 * On Linux, a jiffie represents the time passed in between two timer
 * interrupts. The number of jiffies per second (HZ) varies across the
 * supported platforms. On i386, where HZ is 100, a timer interrupt is
 * generated every 10ms; the resolution is a lot better on ia64, where
 * HZ is 1024. NV_MSECS_TO_JIFFIES should be accurate independent of
 * the actual value of HZ; any partial jiffies will be 'floor'ed, the
 * remainder will be accounted for with mdelay().
 */

RM_STATUS os_delay(U032 MilliSeconds)
{
    unsigned long jiffies;
    unsigned long mdelay_safe_msec;
    
    if (in_irq() && MilliSeconds > NV_MAX_ISR_MDELAY)
        return RM_ERROR;

    if (in_interrupt()) 
    {
        mdelay(MilliSeconds);
        return RM_OK;
    }

    /* do we have a full jiffie to wait? */
    jiffies = NV_MSECS_TO_JIFFIES(MilliSeconds);

    if (jiffies) 
    {
        /* give up the cpu */
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(jiffies);
    }

    /* catch the remainder, if any */
    if (NV_MSECS_PER_JIFFIE) 
    {
        mdelay_safe_msec = MilliSeconds % NV_MSECS_PER_JIFFIE;
        if (mdelay_safe_msec)
            mdelay(mdelay_safe_msec);
    }

    return RM_OK;
}

/* return CPU frequency in MHz */
U032 os_get_cpu_frequency(VOID)
{
    u64 tsc[2];
    u32 tsc_d;
    u32 cpu_mhz;

    tsc[0] = nv_rdtsc();
    mdelay(250);
    tsc[1] = nv_rdtsc();

    /* we timed 1/4th a second */
    tsc_d = (tsc[1] - tsc[0]) * 4;

    /* convert from Hz to MHz */
    cpu_mhz = tsc_d / 1000000;

    return cpu_mhz;
}

RM_STATUS os_get_current_process(U032 *pPid)
{
    *pPid = current->pid;
    return RM_OK;
}

RM_STATUS os_kill_process(
    U032 pid,
    U032 sig
)
{
    return kill_proc(pid, sig, 1) ? RM_ERR_OPERATING_SYSTEM : RM_OK;
}

/*******************************************************************************/
/*                                                                             */
/* Debug and logging utilities follow                                          */
/*                                                                             */
/*******************************************************************************/

#undef DEBUG
#define DEBUG 1         // for rest of this file, DEBUG is always on

/*
 *  print or log error and freakout messages 
 *
 * These routines provide general purpose error reporting.
 * nv_error reports an error to stderr and allows use of
 * printf style formatting.  A newline is appended to all messages.
 * 
 * error_flag can be specified as any of the following:
 *
 *      NV_ERROR_ERRNO       -- include errno text in output
 *      NV_ERROR_STATUS      -- RM status is in error_flag
 *      NV_ERROR_PANIC       -- halts local system after output
 *      NV_ERROR_MESSAGE     -- just do the printing; no hint of error
 *
 *
 * It can also include an errno value which can be OR'd
 * with the above flags.  In this case, the errno within error_flag
 * is used instead of the global errno.
 * 
 * 
 * EXAMPLE
 *  #include <nv.h>
 *  nv_error(0, "stray interrupt %d", intr);
 *
 * EXAMPLE
 *        if ((fd = open(pathname, O_RDNLY)) < 0)
 *        {
 *            nv_error(NV_ERROR_ERRNO, "open of '%s' failed", pathname);
 *            goto failed;
 *        }
 *
 */

//---------------------------------------------------------------------------
//
//  Debugging support.
//
//---------------------------------------------------------------------------

//
// this is what actually outputs the data.
//
inline void out_string(const char *str)
{
    printk("%s", str);
}    

const char *nv_rm_status_text(
    int local_errno
)
{
    static char buff[32];
    static struct {
        unsigned int key;
        char *text;
    } *ep, errs[] = {
    { RM_OK,                            "RM_OK" },
    { RM_ERROR,                         "RM_ERROR" },
    { NV_TYPE_CONFLICT,                 "NV_TYPE_CONFLICT" },
    { NV_OUT_OF_RANGE,                  "NV_OUT_OF_RANGE" },
    { NV_NO_CONNECTION,                 "NV_NO_CONNECTION" },
    { NV_NO_SUCH_OBJECT,                "NV_NO_SUCH_OBJECT" },
    { NV_NAME_IN_USE,                   "NV_NAME_IN_USE" },
    { NV_OUT_OF_RESOURCES,              "NV_OUT_OF_RESOURCES" },
    { NV_TRANSLATION_VIOLATION,         "NV_TRANSLATION_VIOLATION" },
    { NV_PROTECTION_VIOLATION,          "NV_PROTECTION_VIOLATION" },
    { NV_BUFFER_BUSY,                   "NV_BUFFER_BUSY" },
    { NV_ILLEGAL_ACCESS,                "NV_ILLEGAL_ACCESS" },
    { NV_BAD_COLORMAP_FORMAT,           "NV_BAD_COLORMAP_FORMAT" },
    { NV_BAD_COLOR_FORMAT,              "NV_BAD_COLOR_FORMAT" },
    { NV_BAD_MONOCHROME_FORMAT,         "NV_BAD_MONOCHROME_FORMAT" },
    { NV_BAD_PATTERN_SHAPE,             "NV_BAD_PATTERN_SHAPE" },
    { NV_BAD_SUBDIVIDE,                 "NV_BAD_SUBDIVIDE" },
    { NV_NO_CURRENT_POINT,              "NV_NO_CURRENT_POINT" },
    { NV_NO_DMA_TRANSLATION,            "NV_NO_DMA_TRANSLATION" },
    { NV_INCOMPLETE_METHOD,             "NV_INCOMPLETE_METHOD" },
    { NV_RESERVED_ADDRESS,              "NV_RESERVED_ADDRESS" },
    { NV_UNIMPLEMENTED_PATCH,           "NV_UNIMPLEMENTED_PATCH" },
    { NV_OS_NAME_ERROR,                 "NV_OS_NAME_ERROR" },
    { NV_INCOMPLETE_PATCH,              "NV_INCOMPLETE_PATCH" },
    { NV_BUFFERGAP_ERROR,               "NV_BUFFERGAP_ERROR" },
    { NV_FIFO_OVERFLOW,                 "NV_FIFO_OVERFLOW" },
    { RM_WARN_NULL_OBJECT,              "RM_WARN_NULL_OBJECT" },
    { RM_WARN_GRAPHICS_DISABLED,        "RM_WARN_GRAPHICS_DISABLED" },
    { RM_ERR_NO_FREE_MEM,               "RM_ERR_NO_FREE_MEM" },
    { RM_ERR_MEM_NOT_FREED,             "RM_ERR_MEM_NOT_FREED" },
    { RM_ERR_PAGE_TABLE_NOT_AVAIL,      "RM_ERR_PAGE_TABLE_NOT_AVAIL" },
    { RM_ERR_NO_FREE_FIFOS,             "RM_ERR_NO_FREE_FIFOS" },
    { RM_ERR_CANT_CREATE_CLASS_OBJS,    "RM_ERR_CANT_CREATE_CLASS_OBJS" },
    { RM_ERR_BAD_OBJECT,                "RM_ERR_BAD_OBJECT" },
    { RM_ERR_INSERT_DUPLICATE_NAME,     "RM_ERR_INSERT_DUPLICATE_NAME" },
    { RM_ERR_OBJECT_NOT_FOUND,          "RM_ERR_OBJECT_NOT_FOUND" },
    { RM_ERR_CREATE_BAD_CLASS,          "RM_ERR_CREATE_BAD_CLASS" },
    { RM_ERR_DELETE_BAD_CLASS,          "RM_ERR_DELETE_BAD_CLASS" },
    { RM_ERR_FIFO_RUNOUT_OVERFLOW,      "RM_ERR_FIFO_RUNOUT_OVERFLOW" },
    { RM_ERR_FIFO_BAD_ACCESS,           "RM_ERR_FIFO_BAD_ACCESS" },
    { RM_ERR_FIFO_OVERFLOW,             "RM_ERR_FIFO_OVERFLOW" },
    { RM_ERR_METHOD_ORDER,              "RM_ERR_METHOD_ORDER" },
    { RM_ERR_METHOD_COUNT,              "RM_ERR_METHOD_COUNT" },
    { RM_ERR_ILLEGAL_OBJECT,            "RM_ERR_ILLEGAL_OBJECT" },
    { RM_ERR_DMA_IN_USE,                "RM_ERR_DMA_IN_USE" },
    { RM_ERR_BAD_DMA_SPECIFIER,         "RM_ERR_BAD_DMA_SPECIFIER" },
    { RM_ERR_INVALID_XLATE,             "RM_ERR_INVALID_XLATE" },
    { RM_ERR_INVALID_START_LENGTH,      "RM_ERR_INVALID_START_LENGTH" },
    { RM_ERR_DMA_MEM_NOT_LOCKED,        "RM_ERR_DMA_MEM_NOT_LOCKED" },
    { RM_ERR_DMA_MEM_NOT_UNLOCKED,      "RM_ERR_DMA_MEM_NOT_UNLOCKED" },
    { RM_ERR_NOTIFY_IN_USE,             "RM_ERR_NOTIFY_IN_USE" },
    { RM_ERR_ILLEGAL_ADDRESS,           "RM_ERR_ILLEGAL_ADDRESS" },
    { RM_ERR_BAD_ADDRESS,               "RM_ERR_BAD_ADDRESS" },
    { RM_ERR_INVALID_COLOR_FORMAT,      "RM_ERR_INVALID_COLOR_FORMAT" },
    { RM_ERR_INVALID_MONO_FORMAT,       "RM_ERR_INVALID_MONO_FORMAT" },
    { RM_ERR_OBJECT_TYPE_MISMATCH,      "RM_ERR_OBJECT_TYPE_MISMATCH" },
    { RM_ERR_INCOMPLETE_PATCH,          "RM_ERR_INCOMPLETE_PATCH" },
    { RM_ERR_INVALID_PATCH,             "RM_ERR_INVALID_PATCH" },
    { RM_ERR_PATCH_TOO_COMPLEX,         "RM_ERR_PATCH_TOO_COMPLEX" },
    { RM_ERR_MAX_PATCH_FANOUT,          "RM_ERR_MAX_PATCH_FANOUT" },
    { RM_ERR_DEVICE_DISCONNECTED,       "RM_ERR_DEVICE_DISCONNECTED" },
#if 0
    /* whence NV_BUFFER_GAP? */
    { RM_ERR_BUFFER_GAP,                "RM_ERR_BUFFER_GAP" },
#endif
    { RM_ERR_INVALID_OS_NAME,           "RM_ERR_INVALID_OS_NAME" },
    { RM_ERR_ILLEGAL_ACTION,            "RM_ERR_ILLEGAL_ACTION" },
    { RM_ERR_ILLEGAL_DIRECTION,         "RM_ERR_ILLEGAL_DIRECTION" },
    { RM_ERR_OUT_OF_TIMER_CALLBACKS,    "RM_ERR_OUT_OF_TIMER_CALLBACKS" },
    { RM_ERR_OPERATING_SYSTEM,          "RM_ERR_OPERATING_SYSTEM" },
    { RM_ERR_BAD_OBJECT_PARENT,         "RM_ERR_BAD_OBJECT_PARENT" },
    { RM_ERR_BAD_OBJECT_HANDLE,         "RM_ERR_BAD_OBJECT_HANDLE" },
    { RM_ERR_OBJECT_IN_USE,             "RM_ERR_OBJECT_IN_USE" },
    { RM_ERR_OBJECT_HAS_CHILDERN,       "RM_ERR_OBJECT_HAS_CHILDERN" },
    { RM_ERR_BAD_CLASS,                 "RM_ERR_BAD_CLASS" },
    { RM_ERR_INSUFFICIENT_RESOURCES,    "RM_ERR_INSUFFICIENT_RESOURCES" },
    { RM_ERR_BAD_FLAGS,                 "RM_ERR_BAD_FLAGS" },
    { RM_ERR_BAD_BASE,                  "RM_ERR_BAD_BASE" },
    { RM_ERR_BAD_LIMIT,                 "RM_ERR_BAD_LIMIT" },
    { RM_ERR_PROTECTION_FAULT,          "RM_ERR_PROTECTION_FAULT" },
    { RM_ERR_MULTIPLE_MEMORY_TYPES,     "RM_ERR_MULTIPLE_MEMORY_TYPES" },
    { RM_ERR_BAD_OBJECT_ERROR,          "RM_ERR_BAD_OBJECT_ERROR" },
    { RM_ERR_BAD_OBJECT_BUFFER,         "RM_ERR_BAD_OBJECT_BUFFER" },
    { RM_ERR_BAD_OFFSET,                "RM_ERR_BAD_OFFSET" },
    { 0, 0 },
    };

    for (ep = errs; ep->text; ep++)
        if ((ep->key & ~NV_ERROR_MASK) == local_errno)
            return ep->text;

    sprintf(buff, "unknown error 0x%x", local_errno);
    return buff;
}

// The current debug display level (default to maximum debug level)
int cur_debuglevel = 0xaaaaaaaa;

// override initial debug level from registry
VOID os_dbg_init(void)
{
    U032 new_debuglevel;
    if (RM_OK == os_read_registry_dword((PHWINFO) 0,
                                     "",
                                     "ResmanDebugLevel",
                                     &new_debuglevel))
    {
        if (new_debuglevel != ~0)
            cur_debuglevel = new_debuglevel;
    }
}

/* we used to have a global lock that we had to drop so incoming interrupts
 * wouldn't hang the machine waiting for the lock. In the switch to finer
 * grained locking, I've had to remove that code. We don't know which card has
 * a lock or which threw the breakpoint. I should probably scan the list of
 * nv_state_t's and drop any held locks before throwing this breakpoint.
 */
VOID os_dbg_breakpoint(void)
{
    out_string("Break\n");

#if defined(CONFIG_X86_REMOTE_DEBUG)
     __asm__ __volatile__ ("int $3");
#elif defined(CONFIG_KDB)
    KDB_ENTER();
#else
    out_string("Skipping INT 3, we don't like kernel panics\n");
#endif
}



#define MAX_ERROR_STRING 256
    
/*
 * Report an error using a variable argument pointer.
 * This is the utility routine for nv_error, nv_abort,
 * and nv_panic.
 * error_flag is as above; printf_format is a normal
 * printf(3) format string, with its concommitant arguments.
 * 
 * Returns the number of characters written.
 */

int nv_verror(
    int   error_flag,
    const char *printf_format,
    va_list arglist
)
{
    int  local_errno;
    char nv_error_string[MAX_ERROR_STRING];
    char *p = nv_error_string;
    
    local_errno = error_flag & ~NV_ERROR_MASK;

    p = nv_error_string;

    /*
     * Put "ERROR" on all messages, unless requested not to.
     */
    
    if ( ! (error_flag & NV_ERROR_MESSAGE))
        p += sprintf(p, "%s: ",
                     (error_flag & NV_ERROR_WARNING) ? "WARNING" : "ERROR");

    /*
     * Do the actual print of user data and args
     */
    
    p += vsprintf(p, printf_format, arglist);
    
    /*
     * Add in "errno" or NV status info if requested
     */
    
    if ((error_flag & NV_ERROR_ERRNO) && (local_errno != 0))
        p += sprintf(p, " (errno=%d)\n", local_errno);
    else if (error_flag & NV_ERROR_STATUS)
        p += sprintf(p, " (%s)\n", nv_rm_status_text(local_errno));
    else
    p += sprintf(p, "\n");

    out_string(nv_error_string);

    /*
     * Bug out if requested
     */

    if (error_flag & NV_ERROR_PANIC)
    {
        nv_error(0, "fatal error");
        os_dbg_breakpoint();
    }

    return p - nv_error_string;
}


/*
 * Report an error using nv_verror.
 * 
 * Returns the number of characters written.
 */

int nv_error(
    int   error_flag,
    const char *printf_format,
    ...
  )
{
    va_list arglist;
    int chars_written;

    va_start(arglist, printf_format);
    chars_written = nv_verror(error_flag, printf_format, arglist);
    va_end(arglist);
    
    return chars_written;
}

/*
 * nv_panic is shorthand for nv_error(NV_ERROR_PANIC, ...)
 */

void nv_panic(
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    /* | 1 sets local_errno & makes sure 'exit' gets non-zero value    */
    (void) nv_verror(NV_ERROR_PANIC | 1, printf_format, arglist);
    /* NORETURN */
    for (;;) ;
}

/*
 * nv_warning is shorthand for nv_error(NV_ERROR_WARNING, ...)
 */

void nv_warning(
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    (void) nv_verror(NV_ERROR_WARNING, printf_format, arglist);
    va_end(arglist);
}


/*
 * nv_message is shorthand for nv_error(NV_ERROR_MESSAGE, ...)
 * So it doesn't print the word ERROR on every message.
 */

void nv_message(
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    (void) nv_verror(NV_ERROR_MESSAGE, printf_format, arglist);
    va_end(arglist);
}


/*
 * nv_printf() prints to the "error" log for the driver.
 * Just like printf, adds no newlines or names
 */

int nv_printf(
    const char *printf_format,
    ...
)
{
    char nv_error_string[MAX_ERROR_STRING];
    char *p = nv_error_string;
    int chars_written;
    va_list arglist;

    va_start(arglist, printf_format);
    chars_written = vsprintf(p, printf_format, arglist);
    va_end(arglist);

    out_string(p);

    return chars_written;
}

/*
 * Like nv_printf(), but pays attention to debuglevel
 * 
 * Returns the number of characters written.
 */

int nv_debug(
    int   debuglevel,
    const char *printf_format,
    ...
  )
{
    char nv_error_string[MAX_ERROR_STRING];
    char *p = nv_error_string;
    va_list arglist;
    int chars_written = 0;

    if (debuglevel >= cur_debuglevel)
    {
        va_start(arglist, printf_format);
        chars_written = vsprintf(p, printf_format, arglist);
        va_end(arglist);
        out_string(p);
    }

    return chars_written;
}

BOOL os_pci_device_present(
    U016 vendor,
    U016 device
)
{
    if (pci_find_device(vendor, device, NULL))
        return 1;
    return 0;
}

VOID* os_pci_init_handle(
    U008 bus,
    U008 slot,
    U008 function,
    U016 *vendor,
    U016 *device
)
{
    struct pci_dev *dev;
    dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
    if (!dev)
        return NULL;
    if (vendor) *vendor = dev->vendor;
    if (device) *device = dev->device;
    return (VOID *) dev;
}

U008 os_pci_read_byte(
    VOID *handle,
    U008 offset
)
{
    U008 value;
    pci_read_config_byte( (struct pci_dev *) handle, offset, &value);
    return value;
}

U016 os_pci_read_word(
    VOID *handle,
    U008 offset
)
{
    U016 value;
    pci_read_config_word( (struct pci_dev *) handle, offset, &value);
    return value;
}

U032 os_pci_read_dword(
    VOID *handle,
    U008 offset
) 
{
    U032 value;
    pci_read_config_dword( (struct pci_dev *) handle, offset, (u32 *) &value);
    return value;
}

VOID os_pci_write_byte(
    VOID *handle,
    U008 offset,
    U008 value
)
{
    pci_write_config_byte( (struct pci_dev *) handle, offset, value);
}

VOID os_pci_write_word(
    VOID *handle,
    U008 offset,
    U016 value
)
{
    pci_write_config_word( (struct pci_dev *) handle, offset, value);
}

VOID os_pci_write_dword(
    VOID *handle,
    U008 offset,
    U032 value
)
{
    pci_write_config_dword( (struct pci_dev *) handle, offset, value);
}

VOID os_io_write_byte(
    PHWINFO pdev,
    U032 address,
    U008 value
)
{
    outb(value, address);
}

VOID os_io_write_word(
    PHWINFO pdev,
    U032 address,
    U016 value
)
{
    outw(value, address);
}

VOID os_io_write_dword(
    PHWINFO pdev,
    U032 address,
    U032 value
)
{
    outl(value, address);
}

U008 os_io_read_byte(
    PHWINFO pdev,
    U032 address
)
{
    return inb(address);
}

U016 os_io_read_word(
    PHWINFO pdev,
    U032 address
)
{
    return inw(address);
}

U032 os_io_read_dword(
    PHWINFO pdev,
    U032 address
)
{
    return inl(address);
}

ULONG os_cli(ULONG flags)
{
    save_flags(flags);
    cli();
    return flags;
}

ULONG os_sti(ULONG flags)
{
    restore_flags(flags);
    return flags;
}


typedef struct wb_mem_t {
    unsigned int start;
    unsigned int size;
    struct wb_mem_t *next;
} wb_mem;

static wb_mem *wb_list = NULL;

int save_wb(unsigned int start, unsigned int size)
{
#ifdef CONFIG_MTRR
    struct wb_mem_t *ptr;

    NV_KMALLOC(ptr, sizeof(struct wb_mem_t));
    if (ptr == NULL)
        return 0;

    ptr->start = start;
    ptr->size  = size;
    ptr->next  = wb_list;
    wb_list    = ptr;

    return 1;
#endif
    return 0;
}

int del_wb(unsigned int start, unsigned int size)
{
#ifdef CONFIG_MTRR
    struct wb_mem_t *ptr;
    struct wb_mem_t *prev;

    prev = ptr = wb_list;

    while (ptr != NULL) {
        if (ptr->start == start && ptr->size == size) {
            if (ptr == wb_list)
                wb_list = ptr->next;
            else
                prev->next = ptr->next;

            NV_KFREE(ptr);
            return 1;
        }
        prev = ptr;
        ptr = prev->next;
    }
#endif
    return 0;
}

/*
 * Don't assume MTRR-specific, as we could add support for PATs to 
 * achieve the same results on a PIII or higher
 */

RM_STATUS os_set_mem_range(
    U032 start,
    U032 size,
    U032 mode
)
{
#ifdef CONFIG_MTRR
    int err;

    /* for now, we only set memory to write-combined */
    if (mode != NV_MEMORY_WRITECOMBINED)
        return RM_ERROR;

    err = mtrr_add(start, size, MTRR_TYPE_WRCOMB, 0x0);
    if (err < 0)
    {
        printk("nvidia: cannot write-combine 0x%08x, %dM\n",
            start, size / (1024 * 1024));
        return RM_ERROR;
    }

    save_wb(start, size);

    return RM_OK;
#endif
    return RM_ERROR;
}

RM_STATUS os_unset_mem_range(
    U032 start,
    U032 size
)
{
#ifdef CONFIG_MTRR
    if (del_wb(start, size))
    {
        mtrr_del(-1, start, size);
    }
    return RM_OK;
#endif
    return RM_ERROR;
}


/*
 * This needs to remap the AGP Aperture from IO space to kernel space,
 * or at least return a pointer to a kernel space mapping of the Aperture
 * should this also check for Write-Combining??
 */

VOID *os_map_kernel_space(
    U032 start,
    U032 size_bytes,
    U032 mode
)
{
    void *vaddr;

    if (in_interrupt())
    {
        nv_printf("trying to map 0x%x to kernel space in interrupt!\n", start);
        os_dbg_breakpoint();
        return NULL;
    }

    {
#if defined (KERNEL_2_2)
        unsigned long map_nr = MAP_NR(__va(start));
        if (map_nr < max_mapnr) {
            vaddr = __va(start);
        } else 
#endif
        {
            if (mode == NV_MEMORY_DEFAULT) {
                 vaddr = ioremap(start, size_bytes);
            } else {
                 vaddr = ioremap_nocache(start, size_bytes);
            }
        }
    }

#ifdef DEBUG
    if (mode == NV_MEMORY_WRITECOMBINED) {
        printk("nvidia: request to write-combine remapped memory!");
        os_dbg_breakpoint();
    }
    else if (mode != NV_MEMORY_UNCACHED) {
        printk("nvidia: request to cache remapped memory!");
        os_dbg_breakpoint();
    }
#endif 

    return vaddr;
}

VOID os_unmap_kernel_space(
    VOID *addr,
    U032 size_bytes
)
{
#if defined (KERNEL_2_2)
    if (MAP_NR(addr) < max_mapnr) {
        // if we didn't want the memory cached, this isn't necessary
        // but we shouldn't be in a timing critical piece of code.
        asm volatile("wbinvd":::"memory");
    } else 
#endif
    {
        iounmap(addr);
    }
}

VOID* os_map_user_space(
    VOID *kaddr,
    VOID **priv,
    U032 size_bytes,
    U032 mode
)
{
    int err;
    void *paddr = NULL, *uaddr = NULL;

    if ( (size_t) kaddr > VMALLOC_START) {
        nv_ioctl_memory_vtop_t parms;
        parms.buffer = (NvP64)(NV_UINTPTR_T) kaddr;
        parms.byte_length = 1;
        parms.physical_addresses = (NvP64)(NV_UINTPTR_T) &paddr;
        nv_vtop(&parms, 0);
    } else {
        paddr = (void *) virt_to_phys(kaddr);
    }

    uaddr = *priv;

    /* finally, let's do it! */
    err = remap_page_range( (size_t) uaddr, (size_t) paddr, size_bytes, 
                            PAGE_SHARED);

    if (err != 0)
    {
        return (void *) NULL;
    }

    return uaddr;
}

VOID os_unmap_user_space(
    VOID *uaddr,
    VOID *priv
)
{
    // I don't think I need to do anything here...
}

VOID* os_map_io_space(
    U032 start,
    U032 size_bytes,
    VOID **priv,
    U032 user,
    U032 mode
)
{
    int err;
    void *uaddr = NULL;

    if (!user)
        return NULL;

    uaddr = *priv;

    /* finally, let's do it! */
    err = remap_page_range( (size_t) uaddr, (size_t) start, size_bytes, 
                            PAGE_SHARED);

    if (err != 0)
    {
        return (void *) NULL;
    }

    return uaddr;
}

VOID os_unmap_io_space(
    VOID *addr,
    U032 size_bytes,
    VOID *priv,
    U032 user
)
{
}

NvPhysAddr os_get_kern_phys_addr(VOID *vaddr)
{
    U032 ret = 0UL;

    if ( (size_t) vaddr > VMALLOC_START) {
        nv_ioctl_memory_vtop_t parms;
        parms.buffer = (NvP64)(NV_UINTPTR_T) vaddr;
        parms.byte_length = 1;
        parms.physical_addresses = (NvP64)(NV_UINTPTR_T) &ret;
        nv_vtop(&parms, 0);
    } else {
        ret = virt_to_phys(vaddr);
    }

    return ret;
}

/* AGPGART support - mostly by Christian Zander [phoenix@minion.de] */

#ifdef AGPGART

typedef struct {
    agp_memory *ptr;
    int num_pages;
    int offset;
#if defined(NVCPU_IA64)
    struct vm_area_struct *vma;
    void  *handle;
#endif
} agp_priv_data;

typedef struct {
    agp_memory *memory;
    void *aperture;
#ifdef CONFIG_MTRR
    int  mtrr;
#endif
    int  ready;
} agp_gart;

typedef struct {
    int           (*backend_acquire)(void);
    void          (*backend_release)(void);
    void          (*copy_info)(agp_kern_info *);
    agp_memory *  (*allocate_memory)(size_t, unsigned int);
    void          (*free_memory)(agp_memory *);
    int           (*bind_memory)(agp_memory *, off_t);
    int           (*unbind_memory)(agp_memory *);
    void          (*enable)(unsigned int);
} agp_operations_struct;

agp_operations_struct agp_ops;
agp_kern_info         agpinfo;
agp_gart              gart;
#if !defined (KERNEL_2_2)
const drm_agp_t       *drm_agp_p;
#endif

#if defined (KERNEL_2_2)
	#define GET_AGPGART_SYMBOL(sym, sym_string)						\
		sym = (void*) GET_MODULE_SYMBOL(0, sym_string);					\
		if (sym == NULL)								\
		{										\
			printk("NVRM: AGPGART: unable to retrieve symbol %s\n", sym_string);	\
			return 1;								\
		}

	#define AGP_BACKEND_ACQUIRE_SYM   __MODULE_STRING(agp_backend_acquire)
	#define AGP_BACKEND_RELEASE_SYM   __MODULE_STRING(agp_backend_release)
	#define AGP_COPY_INFO_SYM         __MODULE_STRING(agp_copy_info)
	#define AGP_ALLOCATE_MEMORY_SYM   __MODULE_STRING(agp_allocate_memory)
	#define AGP_FREE_MEMORY_SYM       __MODULE_STRING(agp_free_memory)
	#define AGP_BIND_MEMORY_SYM       __MODULE_STRING(agp_bind_memory)
	#define AGP_UNBIND_MEMORY_SYM     __MODULE_STRING(agp_unbind_memory)
	#define AGP_ENABLE_SYM            __MODULE_STRING(agp_enable)
#endif

BOOL KernInitAGP(VOID **ap_phys_base, VOID **ap_mapped_base, U032 *apsize)
{
    U032  agp_rate;
    U032  agp_sba;
    U032  agp_fw;
    char* chipset;

    memset( (void *) &gart, 0, sizeof(agp_gart));

#if !defined (KERNEL_2_2)
    if (!(drm_agp_p = inter_module_get_request("drm_agp", "agpgart"))) {
        printk(KERN_ERR "NVRM: AGPGART: unable to retrieve symbol table\n");
        return 1;
    }
    
    agp_ops.backend_acquire = drm_agp_p->acquire;
    agp_ops.backend_release = drm_agp_p->release;
    agp_ops.copy_info       = drm_agp_p->copy_info;
    agp_ops.allocate_memory = drm_agp_p->allocate_memory;
    agp_ops.free_memory     = drm_agp_p->free_memory;
    agp_ops.bind_memory     = drm_agp_p->bind_memory;
    agp_ops.unbind_memory   = drm_agp_p->unbind_memory;
    agp_ops.enable          = drm_agp_p->enable;

#else
	#if defined(CONFIG_KMOD)
		if ( request_module("agpgart") ) {
			printk(KERN_INFO "NVRM: AGPGART: not loading agpgart.o\n");
			return 1;
		}
	#endif

    GET_AGPGART_SYMBOL(agp_ops.backend_acquire, AGP_BACKEND_ACQUIRE_SYM);
    GET_AGPGART_SYMBOL(agp_ops.backend_release, AGP_BACKEND_RELEASE_SYM);
    GET_AGPGART_SYMBOL(agp_ops.copy_info,       AGP_COPY_INFO_SYM);
    GET_AGPGART_SYMBOL(agp_ops.allocate_memory, AGP_ALLOCATE_MEMORY_SYM);
    GET_AGPGART_SYMBOL(agp_ops.free_memory,     AGP_FREE_MEMORY_SYM);
    GET_AGPGART_SYMBOL(agp_ops.bind_memory,     AGP_BIND_MEMORY_SYM);
    GET_AGPGART_SYMBOL(agp_ops.unbind_memory,   AGP_UNBIND_MEMORY_SYM);
    GET_AGPGART_SYMBOL(agp_ops.enable,          AGP_ENABLE_SYM);
#endif

    /* NOTE: from here down, return an error code of '-1'
     * that indicates that agpgart is loaded, but we failed to use it
     * in some way. This is so we don't try to use nvagp and lock up
     * the memory controller.
     */

    if ( (*(agp_ops.backend_acquire))() ) {
        printk(KERN_ERR "NVRM: AGPGART: backend in use\n");
        return -1;
    }

    if (os_read_registry_dword(NULL, "NVreg", "ReqAGPRate", &agp_rate) == RM_ERROR)
        agp_rate = 0x00000007;
    agp_rate &= 0x00000007;

    if (os_read_registry_dword(NULL, "NVreg", "EnableAGPSBA", &agp_sba) == RM_ERROR)
        agp_sba = 1;
    agp_sba &= 0x00000001;

    if (os_read_registry_dword(NULL, "NVreg", "EnableAGPFW", &agp_fw) == RM_ERROR)
        agp_fw = 1;
    agp_fw &= 0x00000001;
    
    (*(agp_ops.copy_info))(&agpinfo);

    switch ( agpinfo.chipset ) {
        case INTEL_GENERIC:    chipset = "Intel";            break;
        case INTEL_LX:         chipset = "Intel 440LX";      break;
        case INTEL_BX:         chipset = "Intel 440BX";      break;
        case INTEL_GX:         chipset = "Intel 440GX";      break;
        case INTEL_I810:       chipset = "Intel i810";       break;
        case INTEL_I840:       chipset = "Intel i840";       break;
#if !defined (KERNEL_2_2)
        case INTEL_I815:       chipset = "Intel i815";       break;
#if !defined(__rh_config_h__)
        case INTEL_I850:       chipset = "Intel i850";       break;
#endif
#endif
#if defined(NVCPU_IA64)
        case INTEL_460GX:      chipset = "Intel 460GX";      break;
#endif
        case VIA_GENERIC:      chipset = "VIA";              break;
        case VIA_VP3:          chipset = "VIA VP3";          break;
        case VIA_MVP3:         chipset = "VIA MVP3";         break;
        case VIA_MVP4:         chipset = "VIA MVP4";         break;
#if !defined (KERNEL_2_2)
        case VIA_APOLLO_KX133: chipset = "VIA Apollo KX133"; break;
        case VIA_APOLLO_KT133: chipset = "VIA Apollo KT133"; break;
#endif
        case VIA_APOLLO_PRO:   chipset = "VIA Apollo Pro";   break;
        case SIS_GENERIC:      chipset = "SiS";              break;
        case AMD_GENERIC:      chipset = "AMD";              break;
        case AMD_IRONGATE:     chipset = "AMD Irongate";     break;
        case ALI_M1541:        chipset = "ALi M1541";        break;
        case ALI_GENERIC:      chipset = "ALi";              break;
        case NOT_SUPPORTED:    chipset = "unsupported";      break;
        default:               chipset = "unknown";
    }

    printk(KERN_INFO "NVRM: AGPGART: %s chipset\n", chipset);

#ifdef CONFIG_MTRR
    if ((gart.mtrr = mtrr_add(agpinfo.aper_base,
                              agpinfo.aper_size * 0x100000,
                              MTRR_TYPE_WRCOMB, 0x0)) < 0)
    {
        /*
         * Failure to set a write-combining range on the AGP aperture may
         * be due to the presence of other memory ranges with conflicting
         * caching  attributes. Play safe and fail AGP initialization.
         */
        printk(KERN_ERR "NVRM: AGPGART: unable to set MTRR write-combining\n");
        return -1;
    }
#endif

    /*
     * Just map a single page for drivers to osGetAddressInfo() will work.
     * This will allow a client to create a DMA context for the entire
     * aperture without mapping the entire aperture.
     */

    if (!(gart.aperture = (void*) ioremap(agpinfo.aper_base, RM_PAGE_SIZE))) {
        printk(KERN_ERR "NVRM: AGPGART: unable to remap aperture\n");
        return -1;
    }

    printk(KERN_INFO "NVRM: AGPGART: aperture: %ldM @ 0x%08lx\n", (unsigned long)agpinfo.aper_size, (unsigned long)agpinfo.aper_base);
    printk(KERN_INFO "NVRM: AGPGART: aperture mapped from 0x%08lx to 0x%08lx\n", agpinfo.aper_base, (unsigned long) gart.aperture);

    if (!agp_sba) agpinfo.mode &= ~0x00000200;
    if (!agp_fw)  agpinfo.mode &= ~0x00000010;

    if (!(agp_rate & 0x00000004)) agpinfo.mode &= ~0x00000004;
    if (!(agp_rate & 0x00000002)) agpinfo.mode &= ~0x00000002;
    
    (*(agp_ops.enable))(agpinfo.mode);
    
    if (agpinfo.mode & 0x00000200)
        printk(KERN_INFO "NVRM: AGPGART: backend supports sba\n");
    if (agpinfo.mode & 0x00000010)
        printk(KERN_INFO "NVRM: AGPGART: backend supports fw\n");
    if (agpinfo.mode & 0x00000004)
        printk(KERN_INFO "NVRM: AGPGART: mode 4x\n");
    else if (agpinfo.mode & 0x00000002)
        printk(KERN_INFO "NVRM: AGPGART: mode 2x\n");
    else if (agpinfo.mode & 0x00000001)
        printk(KERN_INFO "NVRM: AGPGART: mode 1x\n");

    *ap_phys_base   = (void*) agpinfo.aper_base;
    *ap_mapped_base = (void*) gart.aperture;
    *apsize         = (agpinfo.aper_size * 0x100000) - 1;

    gart.ready = 1;

    return 0;
}

BOOL KernTeardownAGP()
{
    /* sanity check to make sure we should actually be here. */
    if (!gart.ready)
    {
        printk(KERN_INFO "NVRM: AGPGART: not ready so not releasing backend\n");
        return 0;
    }

    if (gart.aperture != NULL) {
#ifdef CONFIG_MTRR
        if (gart.mtrr > 0)
            mtrr_del(gart.mtrr, 0, 0);
#endif
        iounmap(gart.aperture);
    }

    (*(agp_ops.backend_release))();

#if !defined (KERNEL_2_2)
    inter_module_put("drm_agp");
#endif

    printk(KERN_INFO "NVRM: AGPGART: backend released\n");
    return 0;
}

#define DEBUG_AGPGART_ALLOCATIONS 0

#if DEBUG_AGPGART_ALLOCATIONS
static void
print_agp_table()
{
    agp_memory *ptr;

    printk("AGP Table:\n"); 
    for (ptr = gart.memory; ptr; ptr = ptr->next)
    {
        printk("  0x%04x - 0x%04x (%d): 0x%p\n", 
                   (unsigned int) ptr->pg_start, 
                   (unsigned int) ptr->pg_start + ptr->page_count,
                   ptr->page_count,
                   ptr->memory);
    } 
}
#endif

#if DEBUG_AGPGART_ALLOCATIONS
#define PRINT_AGPGART_TABLE() print_agp_table()
#else
#define PRINT_AGPGART_TABLE()
#endif

RM_STATUS KernAllocAGPPages(VOID **pAddress, U032 PageCount,
                            VOID **pPriv_data)
{
    agp_memory *ptr, *i;
    off_t offset;
    int err;
    agp_priv_data *data;
    RM_STATUS status;

    if (!gart.ready) {
        *pAddress = (void*) 0;
        printk(KERN_ERR "NVRM: AGPGART: backend not ready\n");
        return RM_ERROR;
    }
    
    ptr = (*agp_ops.allocate_memory)(PageCount, AGP_NORMAL_MEMORY);
    if (ptr == NULL) {
        *pAddress = (void*) 0;
        printk(KERN_ERR "NVRM: AGPGART: no pages available\n");
        return RM_ERR_NO_FREE_MEM;
    }
    
    offset = 0;

    if (gart.memory == NULL) {
        ptr->next = NULL;
        ptr->prev = NULL;
        gart.memory = ptr;
    } else {
        for (i = gart.memory; i != NULL; i = i->next) {
            if (i->next == NULL && offset == 0) {
                ptr->next = NULL;
                ptr->prev = i;
                i->next = ptr;
                offset = i->pg_start + i->page_count;
                break;
            }
            if (i->pg_start + i->page_count + PageCount < i->next->pg_start) {
                ptr->next = i->next;
                ptr->prev = i;
                i->next->prev = ptr;
                i->next = ptr;
                offset = i->pg_start + i->page_count;
                break;
            }
        }
    }

    err = (*(agp_ops.bind_memory))(ptr, offset);
    if (err) {
        printk(KERN_ERR "NVRM: AGPGART: unable to bind %lu pages\n",
               (unsigned long) PageCount);
        goto fail;
    }

    PRINT_AGPGART_TABLE();
  
    /* return the agp aperture address */ 
    *pAddress = (void *) (agpinfo.aper_base + (offset << PAGE_SHIFT));

    status = os_alloc_mem((VOID **)&data, sizeof(agp_priv_data));
    if (status != RM_OK) {
        printk(KERN_ERR "NVRM: memory allocation failed\n");
        (*(agp_ops.unbind_memory))(ptr);
        goto fail;
    }

    data->ptr = ptr;
    data->num_pages = PageCount;
    data->offset = offset;

    *pPriv_data = data;
    
    printk(KERN_INFO "NVRM: AGPGART: allocated %lu pages\n",
           (unsigned long) PageCount);
    return RM_OK;

fail:
    if (!(ptr->prev == NULL)) {
        if (!(ptr->next == NULL)) {
            ptr->prev->next = ptr->next;
            ptr->next->prev = ptr->prev;
        } else ptr->prev->next = NULL;
    } else gart.memory = ptr->next;

    (*(agp_ops.free_memory))(ptr);
    *pAddress = (void*) 0;

    return RM_ERROR;
}

#if !defined(NVCPU_IA64) || (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))

RM_STATUS 
KernMapAGPPages(void *pvma, void *priv_data)
{
    struct vm_area_struct *vma = (struct vm_area_struct *) pvma;
    agp_priv_data *agp_data = (agp_priv_data *) priv_data;
    unsigned long agp_addr;
    int err;

    if ((agp_data == NULL) || (vma == NULL))
        return RM_ERROR;

    agp_addr = agpinfo.aper_base + (agp_data->offset << PAGE_SHIFT);

    err = remap_page_range(vma->vm_start, (size_t) agp_addr, 
                           agp_data->num_pages << PAGE_SHIFT,
#if defined(NVCPU_IA64)
                           vma->vm_page_prot);
#else
                           PAGE_SHARED);
#endif
        
    if (err) {
        printk(KERN_ERR "NVRM: AGPGART: unable to remap %lu pages\n",
               (unsigned long)agp_data->num_pages);
        (*(agp_ops.unbind_memory))(agp_data->ptr);
        goto fail;
    }
    
#if defined(NVCPU_IA64)
    vma->vm_pgoff = agp_addr >> PAGE_SHIFT;
    agp_data->vma = vma;
    agp_register_map(vma);
#endif

    return RM_OK;

fail:
    /* XXX what else needs to happen here? */
    return RM_ERROR;
}

#else

RM_STATUS 
KernMapAGPNopage(void *address, void *pvma, void *priv_data, void **ppage)
{
    struct vm_area_struct *vma = (struct vm_area_struct *) pvma;
    agp_priv_data *agp_data = (agp_priv_data *) priv_data;
    agp_memory *agp_memory_ptr = agp_data->ptr;
    struct page *page_ptr;
    char *cp;
    unsigned long phys_addr;
    int idx, i;

    idx = ((unsigned long)address - vma->vm_start) >> PAGE_SHIFT;
    phys_addr = (agp_memory_ptr->memory[idx] & 0xffffff) << 12;
    page_ptr = virt_to_page(__va(phys_addr));
    get_page(page_ptr);
    *ppage = page_ptr;
    
    /*
     * Reliability is dramatically improved when invalidating the cache lines
     * prior to using write combined mappings.  The "fc" instruction is
     * specified to affect at least 32 bytes, thus the 32 byte increment below.
     */ 
    cp = (char *)__va(phys_addr);
    for (i = 0; i < PAGE_SIZE/32; i++) {
        ia64_fc (cp);
        cp += 32;
    }
    ia64_sync_i(); /* paranoia */

    return RM_OK;
}
#endif

RM_STATUS KernFreeAGPPages(VOID **pAddress, VOID *priv_data)
{
    agp_memory *ptr;
    agp_priv_data *data = priv_data;
    size_t pages;
    
#if defined(NVCPU_IA64) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))
    agp_unregister_map(data->vma);
#endif
    ptr = data->ptr;

    os_free_mem(data);

    if (!(ptr == NULL)) {
        pages = ptr->page_count;

        if (!(ptr->prev == NULL)) {
            if (!(ptr->next == NULL)) {
                ptr->prev->next = ptr->next;
                ptr->next->prev = ptr->prev;
            } else ptr->prev->next = NULL;
        } else {
            gart.memory = ptr->next;
            if (gart.memory)
                gart.memory->prev = NULL;
        }
    
        (*(agp_ops.unbind_memory))(ptr);
        (*(agp_ops.free_memory))(ptr);

        printk(KERN_INFO "NVRM: AGPGART: freed %ld pages\n", (unsigned long)pages);
        PRINT_AGPGART_TABLE();
        return RM_OK;
    }
    return RM_ERROR;
}

#else

BOOL KernInitAGP(VOID **ap_phys_base, VOID **ap_mapped_base, U032 *apsize)
{
    return 1;
};

BOOL KernTeardownAGP()
{
    return 1;
};

RM_STATUS KernAllocAGPPages(VOID **pAddress, U032 PageCount,
                            VOID **pPriv_data)
{
    *pAddress = (void*) 0;
    return RM_ERROR;
};

RM_STATUS KernMapAGPPages(VOID *pvma, VOID *priv_data)
{
    return RM_ERROR;
}

RM_STATUS KernFreeAGPPages(VOID **pAddress, VOID  *priv_data)
{
    return RM_OK;
};

#endif

// SGI SwapGroup/SwapBarrier extension
// This requires special hardware plus an external driver
// XXX - add code to allocate struct as needed, for each barrier id

// #define TESTING_SWAP 1

#if TESTING_SWAP
// testing code, showing how I tested things

#define ADD_BARRIER_FUNC    "sgitest_add_barrier"
#define REMOVE_BARRIER_FUNC "sgitest_remove_barrier"
#define SWAP_READY_FUNC     "sgitest_swap_ready"

// functions added to serial driver in kernel, from shrijeet
extern int sgi_add_barrier(int id);
extern int sgi_remove_barrier(int id);
extern int sgi_swap_ready(int id, void (*callback)(void *), void *data);

int
sgitest_add_barrier(int id)
{
    return sgi_add_barrier(id);
}

int 
sgitest_remove_barrier(int id)
{
    return sgi_remove_barrier(id);
}

int 
sgitest_swap_ready(int id, void (*callback)(void *), void *data)
{
    return sgi_swap_ready(id, callback, data);
}
#else

#define ADD_BARRIER_FUNC    "sgi_add_barrier"
#define REMOVE_BARRIER_FUNC "sgi_remove_barrier"
#define SWAP_READY_FUNC     "sgi_swap_ready"

#endif


struct barrier_funcs {
    int (*add_barrier)(int id);
    int (*remove_barrier)(int id);
    int (*swap_ready)(int id, void (*callback)(void *), void *data);
};

static struct barrier_funcs sgi_funcs = { NULL, NULL, NULL };

RM_STATUS os_init_swap_barrier(U032 id)
{
    // if we haven't been called before, initialize
    if (sgi_funcs.add_barrier == NULL)
    {
#if defined(TESTING_SWAP)
#if !defined (KERNEL_2_2)
        inter_module_register(ADD_BARRIER_FUNC,    THIS_MODULE, sgitest_add_barrier);
        inter_module_register(REMOVE_BARRIER_FUNC, THIS_MODULE, sgitest_remove_barrier);
        inter_module_register(SWAP_READY_FUNC,     THIS_MODULE, sgitest_swap_ready);
#endif
#endif
        sgi_funcs.add_barrier = GET_MODULE_SYMBOL(0, ADD_BARRIER_FUNC);
        sgi_funcs.remove_barrier = GET_MODULE_SYMBOL(0, REMOVE_BARRIER_FUNC);
        sgi_funcs.swap_ready = GET_MODULE_SYMBOL(0, SWAP_READY_FUNC);
    }

    // failed to initialize
    if (sgi_funcs.add_barrier == NULL)
    {
#if defined(TESTING_SWAP)
        printk("FAILED TO INITIALIZE SWAP!!\n");
#endif
        return RM_ERROR;
    }

    return sgi_funcs.add_barrier(id);
}

RM_STATUS os_remove_swap_barrier(U032 id)
{
    RM_STATUS rmStatus;

    if (sgi_funcs.remove_barrier == NULL)
        return RM_ERROR;

    rmStatus = sgi_funcs.remove_barrier(id);

    if (sgi_funcs.add_barrier)
    {
        PUT_MODULE_SYMBOL(ADD_BARRIER_FUNC);
        sgi_funcs.add_barrier = NULL;
    }

    if (sgi_funcs.remove_barrier)
    {
        PUT_MODULE_SYMBOL(REMOVE_BARRIER_FUNC);
        sgi_funcs.remove_barrier = NULL;
    }

    if (sgi_funcs.swap_ready)
    {
        PUT_MODULE_SYMBOL(SWAP_READY_FUNC);
        sgi_funcs.swap_ready = NULL;
    }

    return rmStatus;
}

RM_STATUS os_swap_barrier(
    U032 id,
    VOID (*callback)(VOID *),
    VOID *data
)
{
    (*callback)((VOID *)data);
    return RM_OK;
}



